# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from i8c import compiler
from i8c import runtime
from i8c.runtime.testcase import BaseTestCase
import io
import os
import struct
import subprocess

class SourceReader(io.BytesIO):
    def readline(self):
        line = io.BytesIO.readline(self)
        trim = line.find(b"//")
        if trim >= 0:
            line = line[:trim] + b"\n"
        return line

class TestOutput(runtime.Context):
    def __init__(self, testcase, index, asm):
        runtime.Context.__init__(self)
        self.__set_fileprefix(testcase, index)
        # Store the assembly language we generated
        asmfile = self.fileprefix + ".S"
        with open(asmfile, "wb") as fp:
            fp.write(asm)
        # Assemble it
        objfile = self.fileprefix + ".o"
        subprocess.check_call(["gcc", "-c", asmfile, "-o", objfile])
        # Load the notes from it
        self.import_notes(objfile)
        self.notes = []
        for notes in self.functions.values():
            self.notes.extend(notes)
        # Make sure we got at least one note
        testcase.assertGreaterEqual(len(self.notes), 1)
        # Setup for note execution
        self.env = testcase
        self.memory = testcase.memory
        self.memory.env = self
        self.register_symbol = testcase.register_symbol

    def __set_fileprefix(self, testcase, index):
        test_id = testcase.id().split(".")
        # Remove the common prefix and the name of the class
        assert test_id[0] == "tests"
        test_id.pop(0)
        test_id.pop(-2)
        # Build the result
        index = "_%04d" % index
        self.fileprefix = os.path.join(
            testcase.topdir, "tests", "output", *test_id) + index
        # Ensure the directory we'll write to exists
        dir = os.path.dirname(self.fileprefix)
        if not os.path.exists(dir):
            os.makedirs(dir)

    @property
    def note(self):
        assert len(self.notes) == 1
        return self.notes[0]

    @property
    def ops(self):
        ops = sorted(self.note.ops.items())
        return [op for pc, op in ops]

    @property
    def opnames(self):
        return [op.name for op in self.ops]

class TestCase(BaseTestCase):
    def __locate_topdir(self):
        self.topdir = os.path.realpath(__file__)
        self.topdir, check = os.path.split(self.topdir)
        assert check.startswith("__init__.py")
        self.topdir, check = os.path.split(self.topdir)
        assert check == "tests"
        assert os.path.exists(os.path.join(self.topdir, "setup.py"))

    def run(self, *args, **kwargs):
        self.__locate_topdir()
        self.compilecount = 0
        return BaseTestCase.run(self, *args, **kwargs)

    def compile(self, input):
        self.compilecount += 1
        input = SourceReader(b'# 1 "<testcase>"\n' + input.encode("utf-8"))
        output = io.BytesIO()
        tree = compiler.compile(input.readline, output.write)
        return tree, TestOutput(self, self.compilecount, output.getvalue())

    def disable_loggers(self):
        for logger in compiler.loggers.values():
            logger.disable()



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import StackError

# Stack underflow checks were off by one.

SOURCE = """\
define test::i8cfail_0002 returns int
    argument int x
    load 5
    blt label
label:
"""

class TestI8CFail0002(TestCase):
    def test_i8cfail_0002(self):
        """Miscellaneous I8C failure #0002 check"""
        self.assertRaises(StackError, self.compile, SOURCE)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import ParserError
from i8c.compiler import StackError
from i8c.compiler import UndefinedIdentError

SOURCE = """\
typedef func ptr (int) derived_func_t
typedef ptr derived_ptr_t
typedef int derived_int_t
typedef opaque derived_opaque_t

define test::test_cast returns ptr
   argument int an_argument
"""

class TestCast(TestCase):
    def test_nocast_fails(self):
        """Check that the test code fails without a cast."""
        self.assertRaises(StackError, self.compile, SOURCE)

    def test_cast(self):
        """Check that cast works."""
        for type in ("int",              # cast to same type
                     "bool",             # upcast to builtin derived type
                     "derived_int_t",    # upcast to user-defined derived type
                     "s32",              # upcast to builtin sized type
                     "ptr",              # cast to unrelated builtin type
                     "derived_ptr_t",    # cast to unrelated derived type
                     "func ptr (int)",   # cast to function
                     "derived_func_t",   # cast to derived function type
                     "opaque",           # cast to opaque
                     "derived_opaque_t", # cast to derived opaque type
                     ):
            for slot in (0, "an_argument", 15, "no_such_slot"):
                source = SOURCE + "cast %s, %s" % (slot, type)

                if (type in ("ptr", "derived_ptr_t")
                    and slot in (0, "an_argument")):
                    error = None
                elif type == "func ptr (int)":
                    error = ParserError
                elif slot == "no_such_slot":
                    error = UndefinedIdentError
                else:
                    error = StackError

                if error is None:
                    tree, output = self.compile(source)
                else:
                    self.assertRaises(error, self.compile, source)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import integer
from .. import constants
from . import ELFFileError
import struct
import sys
import weakref

class ELFFile(object):
    ELFCLASS32 = 1
    ELFCLASS64 = 2
    WORDSIZES = {ELFCLASS32: 32, ELFCLASS64: 64}

    ELFDATA2LSB = 1
    ELFDATA2MSB = 2
    BYTEORDERS = {ELFDATA2LSB: b"<", ELFDATA2MSB: b">"}

    __open = open

    def __init__(self, filename):
        self.filename = filename
        with self.__open(self.filename, "rb") as fp:
            self.bytes = fp.read()
        self.start, self.limit = 0, len(self.bytes)
        hdrfmt = b"4sBB"
        hdrlen = struct.calcsize(hdrfmt)
        magic, ei_class, ei_data = struct.unpack(hdrfmt, self.bytes[:hdrlen])
        if magic != b"\x7fELF":
            raise ELFFileError(filename, "not an ELF file")
        try:
            self.wordsize = self.WORDSIZES[ei_class]
            self.byteorder = self.BYTEORDERS[ei_data]
        except KeyError:
            raise ELFFileError(filename, "unhandled ELF file")

    @property
    def infinity_notes(self):
        # XXX this is not a real ELF parser!
        notename = b"GNU\0"
        markerfmt = self.byteorder + ("I%ds" % len(notename)).encode("utf-8")
        marker = struct.pack(markerfmt, constants.NT_GNU_INFINITY, notename)
        hdrfmt = self.byteorder + b"2I"
        start = hdrsz = struct.calcsize(hdrfmt)
        while True:
            start = self.bytes.find(marker, start)
            if start < 0:
                break
            start -= hdrsz
            namesz, descsz = struct.unpack(
                hdrfmt, self.bytes[start:start + hdrsz])
            assert namesz == len(notename)
            descstart = start + hdrsz + struct.calcsize("I") + len(notename)
            desclimit = descstart + descsz
            yield self[descstart:desclimit]
            start = desclimit

    def __getitem__(self, key):
        return ELFSlice(weakref.ref(self), key)

class ELFSlice(object):
    def __init__(self, elffile, ourslice):
        assert isinstance(ourslice, slice)
        assert ourslice.step in (None, 1)
        self.elffile = elffile

        elffile = elffile()
        self.filename = elffile.filename
        self.wordsize = elffile.wordsize
        self.byteorder = elffile.byteorder

        assert ourslice.start >= 0
        self.start = elffile.start + ourslice.start
        assert self.start <= elffile.limit
        assert ourslice.stop >= 0
        self.limit = elffile.start + ourslice.stop
        assert self.limit <= elffile.limit

    def __getitem__(self, key):
        if isinstance(key, integer):
            result = self.bytes[key]
            if isinstance(result, int):
                result = chr(result)
            return result

        assert isinstance(key, slice)
        assert key.step in (None, 1)

        start = self.start
        if key.start is not None:
            start += key.start
            assert start <= self.limit
        if key.stop is not None:
            limit = self.start + key.stop
            assert limit <= self.limit
        else:
            limit = self.limit
        return self.elffile()[start:limit]

    def __len__(self):
        return self.limit - self.start

    def __add__(self, offset):
        start = self.start + offset
        assert start <= self.limit
        return self.elffile()[start:self.limit]

    @property
    def bytes(self):
        return self.elffile().bytes[self.start:self.limit]

    @property
    def text(self):
        text = self.bytes
        if sys.version_info >= (3,):
            text = "".join(map(chr, text))
        assert isinstance(text, str)
        return text

open = ELFFile
